// Driver-Enumeration and handling utility
#define DRIVER_MAX 136

#define DTYP_NOTHING 0
#define DTYP_CHAR 1
#define DTYP_BLOCK 2
#define DTYP_FIXEDDISK 3
#define DTYP_RAMDISK 4
#define DTYP_FILESYSTEM 5
#define DTYP_DISPLAY 6
#define DTYP_MOUSE 7
#define DTYP_BINEMU 8
#define DTYP_WINDOWMANAGER 9
#define DTYP_KEYBOARD 10
#define DTYP_OTHER 255

#pragma pack(1)
struct drvlst
    {
    unsigned char used;
    unsigned char dtyp;
    unsigned int length;   // For Block&Char-Drivers
    void ( __far * drvcall)();
    char *dname;
    } drvlist[DRIVER_MAX];

void drvlist_init()
    {
    drvlist[128].used = 1;
    drvlist[128].dtyp = DTYP_FILESYSTEM;
    drvlist[128].dname = "DOS-FS";
    drvlist[128].length = 0;
    drvlist[128].drvcall = NULL;

    drvlist[129].used = 1;
    drvlist[129].dtyp = DTYP_CHAR;
    drvlist[129].dname = "tty0";
    drvlist[129].length = 1;
    drvlist[129].drvcall = NULL;

    drvlist[130].used = 1;
    drvlist[130].dtyp = DTYP_CHAR;
    drvlist[130].dname = "null";
    drvlist[130].drvcall = NULL;

    drvlist[131].used = 1;
    drvlist[131].dtyp = DTYP_FILESYSTEM;
    drvlist[131].dname = "DEV-FS";
    drvlist[131].drvcall = NULL;

    if(xmscall != NULL)
        {
        drvlist[132].used = 1;
        drvlist[132].dtyp = DTYP_RAMDISK;
        drvlist[132].dname = "xms";
        drvlist[132].drvcall = NULL;
        }

    drvlist[133].used = 1;
    drvlist[133].dtyp = DTYP_WINDOWMANAGER;
    drvlist[133].dname = "ulsopwm";
    drvlist[133].drvcall = NULL;

    drvlist[134].used = 1;
    drvlist[134].dtyp = DTYP_CHAR;
    drvlist[134].dname = "fifo";
    drvlist[134].drvcall = NULL;

    drvlist[135].used = 1;
    drvlist[135].dtyp = DTYP_CHAR;
    drvlist[135].dname = "waitfifo";
    drvlist[135].drvcall = NULL;
    }

// This gets a driver
int getdrvid(char *drvnam)
    {
    int i;
    for(i=0;i<DRIVER_MAX;i++)
        if(drvlist[i].used)
            if(!strcmp(drvlist[i].dname,drvnam))
                return i;
    return ERR_INVALIDDRIVER;
    }

int adddriver(int type, char *name, unsigned int length, void *callback)
    {
    int i;
    for(i=0;i<DRIVER_MAX;i++)
        if(!drvlist[i].used)
            {
            drvlist[i].used = 1;
            drvlist[i].dname = name;
            drvlist[i].dtyp = type;
            drvlist[i].length = length;
            drvlist[i].drvcall = callback;
            return i;
            }
    return ERR_OUTOFDRIVERHANDLES;
    }

int insmod(char *filename)
    {
    int hndl,ii;
    char istr[5];
    UCHR typ;
    UINT numload,i,j;
    ULNG replen;
    char *buf;
    istr[4] = 0;

    if((hndl = fsopen(filename,0)) < 0) return hndl;
    fsread(hndl, 4, istr);
    if(strcmp(istr, "DRV>"))
        {
        fsclose(hndl);
        return ERR_INVALIDEXE;
        }
    fsgetc(hndl);   // Ignore flags
    typ = fsgetc(hndl);
    numload = fsgetw(hndl);
    replen = fsgetdw(hndl);
    if(numload == 0)
        numload = fhandles[hndl].len - fhandles[hndl].pos;
    if((buf = mem_malloc(numload,0)) == NULL)
        {
        fsclose(hndl);
        return ERR_OUTOFMEMORY;
        }
    fsread(hndl,numload,buf);
    fsclose(hndl);

//    printf("Driver loaded @ %p, debugging\n",buf);
//    printf("This function is @ %p\n",insmod);

//    system("debug"); exit(100);

    _asm
        {
        push ds
        xor ax, ax
        call buf
        mov ii, ax
        mov ax, ds
        mov i, ds
        mov j, bx
        pop ds
        }
    if(ii < 0)
        {
        printf("memfreeinsmoderr");
        mem_free(buf);
        return ii;
        }
    if((ii = adddriver(typ,MK_FP(i,j),replen,buf)) < 0)
        {
        // That's bad. The driver already initialised itself and now
        // There's no handle. So deinitialise and remove it
        i = FP_SEG(buf);
        _asm
            {
            push ds
            mov ax, 4
            call buf
            pop ds
            }
        printf("memfreenoinsmodhndl");
        mem_free(buf);
        return ii;
        }
    return 0;
    }
